# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.ml.constraints import ConstraintSchema
from polyaxon_schemas.ml.initializations import (
    GlorotNormalInitializerConfig,
    InitializerSchema,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.ml.regularizations import RegularizerSchema
from polyaxon_schemas.utils import ACTIVATION_VALUES, ObjectOrListObject, StrOrFct

# pylint:disable=too-many-lines


class Conv1DSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=1, max=1)
    strides = ObjectOrListObject(fields.Int, min=1, max=1, default=1, missing=1)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    dilation_rate = ObjectOrListObject(fields.Int, min=1, max=1, default=1, missing=1)
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return Conv1DConfig


class Conv1DConfig(BaseLayerConfig):
    """1D convolution layer (e.g. temporal convolution).

    This layer creates a convolution kernel that is convolved
    with the layer input over a single spatial (or temporal) dimension
    to produce a tensor of outputs.
    If `use_bias` is True, a bias vector is created and added to the outputs.
    Finally, if `activation` is not `None`,
    it is applied to the outputs as well.

    When using this layer as the first layer in a model,
    provide an `input_shape` argument
    (tuple of integers or `None`, e.g.
    `(10, 128)` for sequences of 10 vectors of 128-dimensional vectors,
    or `(None, 128)` for variable-length sequences of 128-dimensional vectors.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of a single integer,
            specifying the length of the 1D convolution window.
        strides: An integer or tuple/list of a single integer,
            specifying the stride length of the convolution.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: One of `"valid"`, `"causal"` or `"same"` (case-insensitive).
            `"causal"` results in causal (dilated) convolutions, e.g. output[t]
            does not depend on input[t+1:]. Useful when modeling temporal data
            where the model should not violate the temporal order.
            See [WaveNet: A Generative Model for Raw Audio, section
              2.1](https://arxiv.org/abs/1609.03499).
        dilation_rate: an integer or tuple/list of a single integer, specifying
            the dilation rate to use for dilated convolution.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any `strides` value != 1.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to the kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        3D tensor with shape: `(batch_size, steps, input_dim)`

    Output shape:
        3D tensor with shape: `(batch_size, new_steps, filters)`
        `steps` value might have changed due to padding or strides.

    Polyaxon usage:

    ```yaml
    Conv1D:
      filters: 10
      kernel_size: 3
      strides: 1
      padding: same
    ```
    """
    IDENTIFIER = 'Conv1D'
    SCHEMA = Conv1DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=1,
                 padding='valid',
                 dilation_rate=1,
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(Conv1DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.dilation_rate = dilation_rate
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class Conv2DSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=2, max=2)
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    dilation_rate = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return Conv2DConfig


class Conv2DConfig(BaseLayerConfig):
    """2D convolution layer (e.g. spatial convolution over images).

    This layer creates a convolution kernel that is convolved
    with the layer input to produce a tensor of
    outputs. If `use_bias` is True,
    a bias vector is created and added to the outputs. Finally, if
    `activation` is not `None`, it is applied to the outputs as well.

    When using this layer as the first layer in a model,
    provide the keyword argument `input_shape`
    (tuple of integers, does not include the sample axis),
    e.g. `input_shape=(128, 128, 3)` for 128x128 RGB pictures
    in `data_format="channels_last"`.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of 2 integers, specifying the
            width and height of the 2D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        strides: An integer or tuple/list of 2 integers,
            specifying the strides of the convolution along the width and height.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: one of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".
        dilation_rate: an integer or tuple/list of 2 integers, specifying
            the dilation rate to use for dilated convolution.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any stride value != 1.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to the kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        4D tensor with shape:
        `(samples, channels, rows, cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(samples, rows, cols, channels)` if data_format='channels_last'.

    Output shape:
        4D tensor with shape:
        `(samples, filters, new_rows, new_cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(samples, new_rows, new_cols, filters)` if data_format='channels_last'.
        `rows` and `cols` values might have changed due to padding.

    Polyaxonfile usage:

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8 or [8, 8]
      strides: 2 or [2, 2]
      padding: valid
      activation: tanh
      kernel_initializer: Ones
    ```

    or


    ```yaml
    Conv2D:
      filters: 10
      kernel_size: [8, 8]
      strides: [2, 2]
      padding: valid
      activation: tanh
      kernel_initializer:
        Ones:
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: [8, 8]
      strides: 1
      padding: valid
      activation: tanh
      kernel_initializer:
        Ones: {dtype: float32}
    ```
    """
    IDENTIFIER = 'Conv2D'
    SCHEMA = Conv2DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 data_format=None,
                 dilation_rate=(1, 1),
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(Conv2DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.dilation_rate = dilation_rate
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class Conv3DSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=3, max=3)
    strides = ObjectOrListObject(fields.Int, min=3, max=3, default=(1, 1, 1), missing=(1, 1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    dilation_rate = ObjectOrListObject(fields.Int, min=3, max=3,
                                       default=(1, 1, 1), missing=(1, 1, 1))
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return Conv3DConfig


class Conv3DConfig(BaseLayerConfig):
    """3D convolution layer (e.g. spatial convolution over volumes).

    This layer creates a convolution kernel that is convolved
    with the layer input to produce a tensor of
    outputs. If `use_bias` is True,
    a bias vector is created and added to the outputs. Finally, if
    `activation` is not `None`, it is applied to the outputs as well.

    When using this layer as the first layer in a model,
    provide the keyword argument `input_shape`
    (tuple of integers, does not include the sample axis),
    e.g. `input_shape=(128, 128, 128, 1)` for 128x128x128 volumes
    with a single channel,
    in `data_format="channels_last"`.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of 3 integers, specifying the
            depth, height and width of the 3D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        strides: An integer or tuple/list of 3 integers,
            specifying the strides of the convolution along each spatial
              dimension.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: one of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".
        dilation_rate: an integer or tuple/list of 3 integers, specifying
            the dilation rate to use for dilated convolution.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any stride value != 1.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to the kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        5D tensor with shape:
        `(samples, channels, conv_dim1, conv_dim2, conv_dim3)` if
          data_format='channels_first'
        or 5D tensor with shape:
        `(samples, conv_dim1, conv_dim2, conv_dim3, channels)` if
          data_format='channels_last'.

    Output shape:
        5D tensor with shape:
        `(samples, filters, new_conv_dim1, new_conv_dim2, new_conv_dim3)` if
          data_format='channels_first'
        or 5D tensor with shape:
        `(samples, new_conv_dim1, new_conv_dim2, new_conv_dim3, filters)` if
          data_format='channels_last'.
        `new_conv_dim1`, `new_conv_dim2` and `new_conv_dim3` values might have
         changed due to padding.

    Polyaxonfile usage:

    ```yaml
    Conv3D:
      filters: 10
      kernel_size: 8 or [8, 8, 8]
      strides: 1 or [1, 1, 1]
      padding: valid
      activation: tanh
      kernel_initializer: Ones
    ```
    """
    IDENTIFIER = 'Conv3D'
    SCHEMA = Conv3DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1, 1),
                 padding='valid',
                 data_format=None,
                 dilation_rate=(1, 1, 1),
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(Conv3DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.dilation_rate = dilation_rate
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class Conv2DTransposeSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=2, max=2)
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    dilation_rate = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return Conv2DTransposeConfig


class Conv2DTransposeConfig(BaseLayerConfig):
    """Transposed convolution layer (sometimes called Deconvolution).

    The need for transposed convolutions generally arises
    from the desire to use a transformation going in the opposite direction
    of a normal convolution, i.e., from something that has the shape of the
    output of some convolution to something that has the shape of its input
    while maintaining a connectivity pattern that is compatible with
    said convolution.

    When using this layer as the first layer in a model,
    provide the keyword argument `input_shape`
    (tuple of integers, does not include the sample axis),
    e.g. `input_shape=(128, 128, 3)` for 128x128 RGB pictures
    in `data_format="channels_last"`.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number of output filters in the convolution).
        kernel_size: An integer or tuple/list of 2 integers, specifying the
            width and height of the 2D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        strides: An integer or tuple/list of 2 integers,
            specifying the strides of the convolution along the width and height.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: one of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".
        dilation_rate: an integer or tuple/list of 2 integers, specifying
            the dilation rate to use for dilated convolution.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any stride value != 1.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to the kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        4D tensor with shape:
        `(batch, channels, rows, cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(batch, rows, cols, channels)` if data_format='channels_last'.

    Output shape:
        4D tensor with shape:
        `(batch, filters, new_rows, new_cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(batch, new_rows, new_cols, filters)` if data_format='channels_last'.
        `rows` and `cols` values might have changed due to padding.

    References:
        - [A guide to convolution arithmetic for deep
          learning](https://arxiv.org/abs/1603.07285v1)
        - [Deconvolutional
          Networks](http://www.matthewzeiler.com/pubs/cvpr2010/cvpr2010.pdf)

    Polyaxonfile usage:

    ```yaml
    Conv2DTranspose:
      filters: 10
      kernel_soze: [4, 4]
    ```
    """
    IDENTIFIER = 'Conv2DTranspose'
    SCHEMA = Conv2DTransposeSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 data_format=None,
                 dilation_rate=(1, 1),
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(Conv2DTransposeConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.activation = activation
        self.dilation_rate = dilation_rate
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class Conv3DTransposeSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=3, max=3)
    strides = ObjectOrListObject(fields.Int, min=3, max=3, default=(1, 1, 1), missing=(1, 1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    dilation_rate = ObjectOrListObject(fields.Int, min=3, max=3,
                                       default=(1, 1, 1), missing=(1, 1, 1))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return Conv3DTransposeConfig


class Conv3DTransposeConfig(BaseLayerConfig):
    """Transposed convolution layer (sometimes called Deconvolution).

    The need for transposed convolutions generally arises
    from the desire to use a transformation going in the opposite direction
    of a normal convolution, i.e., from something that has the shape of the
    output of some convolution to something that has the shape of its input
    while maintaining a connectivity pattern that is compatible with
    said convolution.

    When using this layer as the first layer in a model,
    provide the keyword argument `input_shape`
    (tuple of integers, does not include the sample axis),
    e.g. `input_shape=(128, 128, 128, 3)` for a 128x128x128 volume with 3 channels
    if `data_format="channels_last"`.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number of output filters in the convolution).
        kernel_size: An integer or tuple/list of 3 integers, specifying the
            depth, height and width of the 3D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        strides: An integer or tuple/list of 3 integers,
            specifying the strides of the convolution along the depth, height
              and width.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: one of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, depth, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, depth, height, width)`.
            If you never set it, then it will be "channels_last".
        dilation_rate: an integer or tuple/list of 3 integers, specifying
            the dilation rate to use for dilated convolution.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Currently, specifying any `dilation_rate` value != 1 is
            incompatible with specifying any stride value != 1.
        activation: Activation function to use
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix
        bias_initializer: Initializer for the bias vector
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix
        bias_regularizer: Regularizer function applied to the bias vector
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation").
        kernel_constraint: Constraint function applied to the kernel matrix
        bias_constraint: Constraint function applied to the bias vector

    Input shape:
        5D tensor with shape:
        `(batch, channels, depth, rows, cols)` if data_format='channels_first'
        or 5D tensor with shape:
        `(batch, depth, rows, cols, channels)` if data_format='channels_last'.

    Output shape:
        5D tensor with shape:
        `(batch, filters, new_depth, new_rows, new_cols)` if
          data_format='channels_first'
        or 5D tensor with shape:
        `(batch, new_depth, new_rows, new_cols, filters)` if
          data_format='channels_last'.
        `depth` and `rows` and `cols` values might have changed due to padding.

    References:
        - [A guide to convolution arithmetic for deep
          learning](https://arxiv.org/abs/1603.07285v1)
        - [Deconvolutional
          Networks](http://www.matthewzeiler.com/pubs/cvpr2010/cvpr2010.pdf)

    Polyaxonfile usage:

    ```yaml
    Conv3DTranspose:
      filters: 10
      kernel_size: 3 or [3, 3, 3]
    ```
    """
    IDENTIFIER = 'Conv3DTranspose'
    SCHEMA = Conv3DTransposeSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1, 1),
                 padding='valid',
                 data_format=None,
                 activation=None,
                 dilation_rate=(1, 1, 1),
                 use_bias=True,
                 kernel_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(Conv3DTransposeConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.activation = activation
        self.dilation_rate = dilation_rate
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class SeparableConv2DSchema(BaseLayerSchema):
    filters = fields.Int()
    kernel_size = ObjectOrListObject(fields.Int, min=2, max=2)
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))
    depth_multiplier = fields.Int(default=1, missing=1)
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    depthwise_initializer = fields.Nested(InitializerSchema, allow_none=True)
    pointwise_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    depthwise_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    pointwise_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    depthwise_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    pointwise_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return SeparableConv2DConfig


class SeparableConv2DConfig(BaseLayerConfig):
    """Depthwise separable 2D convolution.

    Separable convolutions consist in first performing
    a depthwise spatial convolution
    (which acts on each input channel separately)
    followed by a pointwise convolution which mixes together the resulting
    output channels. The `depth_multiplier` argument controls how many
    output channels are generated per input channel in the depthwise step.

    Intuitively, separable convolutions can be understood as
    a way to factorize a convolution kernel into two smaller kernels,
    or as an extreme version of an Inception block.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number output of filters in the convolution).
        kernel_size: An integer or tuple/list of 2 integers, specifying the
            width and height of the 2D convolution window.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        strides: An integer or tuple/list of 2 integers,
            specifying the strides of the convolution along the width and height.
            Can be a single integer to specify the same value for
            all spatial dimensions.
            Specifying any stride value != 1 is incompatible with specifying
            any `dilation_rate` value != 1.
        padding: one of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".
        depth_multiplier: The number of depthwise convolution output channels
            for each input channel.
            The total number of depthwise convolution output
            channels will be equal to `filterss_in * depth_multiplier`.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        depthwise_initializer: Initializer for the depthwise kernel matrix.
        pointwise_initializer: Initializer for the pointwise kernel matrix.
        bias_initializer: Initializer for the bias vector.
        depthwise_regularizer: Regularizer function applied to
            the depthwise kernel matrix.
        pointwise_regularizer: Regularizer function applied to
            the pointwise kernel matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        depthwise_constraint: Constraint function applied to
            the depthwise kernel matrix.
        pointwise_constraint: Constraint function applied to
            the pointwise kernel matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        4D tensor with shape:
        `(batch, channels, rows, cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(batch, rows, cols, channels)` if data_format='channels_last'.

    Output shape:
        4D tensor with shape:
        `(batch, filters, new_rows, new_cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(batch, new_rows, new_cols, filters)` if data_format='channels_last'.
        `rows` and `cols` values might have changed due to padding.

    Polyaxonfile usage:

    ```yaml
    SeparableConv2D:
      filters: 10
      kernel_size: 2 or [2, 2]
    ```
    """
    IDENTIFIER = 'SeparableConv2D'
    SCHEMA = SeparableConv2DSchema

    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 data_format=None,
                 depth_multiplier=1,
                 activation=None,
                 use_bias=True,
                 depthwise_initializer=GlorotNormalInitializerConfig(),
                 pointwise_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 depthwise_regularizer=None,
                 pointwise_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 depthwise_constraint=None,
                 pointwise_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(SeparableConv2DConfig, self).__init__(**kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format
        self.depth_multiplier = depth_multiplier
        self.activation = activation
        self.use_bias = use_bias
        self.depthwise_initializer = depthwise_initializer
        self.pointwise_initializer = pointwise_initializer
        self.bias_initializer = bias_initializer
        self.depthwise_regularizer = depthwise_regularizer
        self.pointwise_regularizer = pointwise_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.depthwise_constraint = depthwise_constraint
        self.pointwise_constraint = pointwise_constraint
        self.bias_constraint = bias_constraint


class UpSampling1DSchema(BaseLayerSchema):
    size = ObjectOrListObject(fields.Int, min=2, max=2, default=2, missing=2)

    @staticmethod
    def schema_config():
        return UpSampling1DConfig


class UpSampling1DConfig(BaseLayerConfig):
    """Upsampling layer for 1D inputs.

    Repeats each temporal step `size` times along the time axis.

    Args:
        size: integer. Upsampling factor.

    Input shape:
        3D tensor with shape: `(batch, steps, features)`.

    Output shape:
        3D tensor with shape: `(batch, upsampled_steps, features)`.

    Polyaxonfile usage:

    ```yaml
    UpSampling1D:
    ```

    or

    ```yaml
    UpSampling1D:
        size: 2
    ```

    or

    ```yaml
    UpSampling1D: {size: 2}
    ```
    """
    IDENTIFIER = 'UpSampling1D'
    SCHEMA = UpSampling1DSchema

    def __init__(self, size=2, **kwargs):
        super(UpSampling1DConfig, self).__init__(**kwargs)
        self.size = size


class UpSampling2DSchema(BaseLayerSchema):
    size = ObjectOrListObject(fields.Int, min=2, max=2, default=(2, 2), missing=(2, 2))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return UpSampling2DConfig


class UpSampling2DConfig(BaseLayerConfig):
    """Upsampling layer for 2D inputs.

    Repeats the rows and columns of the data
    by size[0] and size[1] respectively.

    Args:
        size: int, or tuple of 2 integers.
            The upsampling factors for rows and columns.
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, rows, cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, rows, cols)`

    Output shape:
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, upsampled_rows, upsampled_cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, upsampled_rows, upsampled_cols)`

    Polyaxonfile usage:

    ```yaml
    UpSampling2D:
    ```

    or

    ```yaml
    UpSampling2D:
        size: 2 or [2, 2]
    ```
    """
    IDENTIFIER = 'UpSampling2D'
    SCHEMA = UpSampling2DSchema

    def __init__(self, size=(2, 2), data_format=None, **kwargs):
        super(UpSampling2DConfig, self).__init__(**kwargs)
        self.size = size
        self.data_format = data_format


class UpSampling3DSchema(BaseLayerSchema):
    size = ObjectOrListObject(fields.Int, min=3, max=3, default=(2, 2, 2), missing=(2, 2, 2))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return UpSampling3DConfig


class UpSampling3DConfig(BaseLayerConfig):
    """Upsampling layer for 3D inputs.

    Repeats the 1st, 2nd and 3rd dimensions
    of the data by size[0], size[1] and size[2] respectively.

    Args:
        size: int, or tuple of 3 integers.
            The upsampling factors for dim1, dim2 and dim3.
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        5D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, dim1, dim2, dim3, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, dim1, dim2, dim3)`

    Output shape:
        5D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, upsampled_dim1, upsampled_dim2, upsampled_dim3, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, upsampled_dim1, upsampled_dim2, upsampled_dim3)`

    Polyaxonfile usage:

    ```yaml
    UpSampling3D:
    ```

    or

    ```yaml
    UpSampling3D:
        size: 2 or [2, 2, 2]
    ```
    """
    IDENTIFIER = 'UpSampling3D'
    SCHEMA = UpSampling3DSchema

    def __init__(self, size=(2, 2, 2), data_format=None, **kwargs):
        super(UpSampling3DConfig, self).__init__(**kwargs)
        self.size = size
        self.data_format = data_format


class ZeroPadding1DSchema(BaseLayerSchema):
    padding = ObjectOrListObject(fields.Int, min=1, max=1, default=1, missing=1)

    @staticmethod
    def schema_config():
        return ZeroPadding1DConfig


class ZeroPadding1DConfig(BaseLayerConfig):
    """Zero-padding layer for 1D input (e.g. temporal sequence).

    Args:
        padding: int, or tuple of int (length 2), or dictionary.
            - If int:
            How many zeros to add at the beginning and end of
            the padding dimension (axis 1).
            - If tuple of int (length 2):
            How many zeros to add at the beginning and at the end of
            the padding dimension (`(left_pad, right_pad)`).

    Input shape:
        3D tensor with shape `(batch, axis_to_pad, features)`

    Output shape:
        3D tensor with shape `(batch, padded_axis, features)`

    Polyaxonfile usage:

    ```yaml
    ZeroPadding1D:
    ```

    or

    ```yaml
    ZeroPadding1D:
        padding: 1
    ```
    """
    IDENTIFIER = 'ZeroPadding1D'
    SCHEMA = ZeroPadding1DSchema

    def __init__(self, padding=1, **kwargs):
        super(ZeroPadding1DConfig, self).__init__(**kwargs)
        self.padding = padding


class ZeroPadding2DSchema(BaseLayerSchema):
    padding = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return ZeroPadding2DConfig


class ZeroPadding2DConfig(BaseLayerConfig):
    """Zero-padding layer for 2D input (e.g. picture).

    This layer can add rows and columns of zeros
    at the top, bottom, left and right side of an image tensor.

    Args:
        padding: int, or tuple of 2 ints, or tuple of 2 tuples of 2 ints.
            - If int: the same symmetric padding
                is applied to width and height.
            - If tuple of 2 ints:
                interpreted as two different
                symmetric padding values for height and width:
                `(symmetric_height_pad, symmetric_width_pad)`.
            - If tuple of 2 tuples of 2 ints:
                interpreted as
                `((top_pad, bottom_pad), (left_pad, right_pad))`
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, rows, cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, rows, cols)`

    Output shape:
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, padded_rows, padded_cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, padded_rows, padded_cols)`

    Polyaxonfile usage:

    ```yaml
    ZeroPadding2D:
    ```

    or

    ```yaml
    ZeroPadding2D:
        padding: 1 or [1, 1]
    ```
    """
    IDENTIFIER = 'ZeroPadding2D'
    SCHEMA = ZeroPadding2DSchema

    def __init__(self, padding=(1, 1), data_format=None, **kwargs):
        super(ZeroPadding2DConfig, self).__init__(**kwargs)
        self.padding = padding
        self.data_format = data_format


class ZeroPadding3DSchema(BaseLayerSchema):
    padding = ObjectOrListObject(fields.Int, min=3, max=3, default=(1, 1, 1), missing=(1, 1, 1))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return ZeroPadding3DConfig


class ZeroPadding3DConfig(BaseLayerConfig):
    """Zero-padding layer for 3D data (spatial or spatio-temporal).

    Args:
        padding: int, or tuple of 2 ints, or tuple of 2 tuples of 2 ints.
            - If int: the same symmetric padding
                is applied to width and height.
            - If tuple of 2 ints:
                interpreted as two different
                symmetric padding values for height and width:
                `(symmetric_dim1_pad, symmetric_dim2_pad, symmetric_dim3_pad)`.
            - If tuple of 2 tuples of 2 ints:
                interpreted as
                `((left_dim1_pad, right_dim1_pad), (left_dim2_pad,
                  right_dim2_pad), (left_dim3_pad, right_dim3_pad))`
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        5D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, first_axis_to_pad, second_axis_to_pad, third_axis_to_pad,
              depth)`
        - If `data_format` is `"channels_first"`:
            `(batch, depth, first_axis_to_pad, second_axis_to_pad,
              third_axis_to_pad)`

    Output shape:
        5D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, first_padded_axis, second_padded_axis, third_axis_to_pad,
              depth)`
        - If `data_format` is `"channels_first"`:
            `(batch, depth, first_padded_axis, second_padded_axis,
              third_axis_to_pad)`

    Polyaxonfile usage:

    ```yaml
    ZeroPadding3D:
    ```

    or

    ```yaml
    ZeroPadding3D:
        padding: 1 or [1, 1, 1]
    ```
    """
    IDENTIFIER = 'ZeroPadding3D'
    SCHEMA = ZeroPadding3DSchema

    def __init__(self, padding=(1, 1, 1), data_format=None, **kwargs):
        super(ZeroPadding3DConfig, self).__init__(**kwargs)
        self.padding = padding
        self.data_format = data_format


class Cropping1DSchema(BaseLayerSchema):
    cropping = ObjectOrListObject(fields.Int, min=2, max=2, default=(1, 1), missing=(1, 1))

    @staticmethod
    def schema_config():
        return Cropping1DConfig


class Cropping1DConfig(BaseLayerConfig):
    """Cropping layer for 1D input (e.g. temporal sequence).

    It crops along the time dimension (axis 1).

    Args:
        cropping: int or tuple of int (length 2)
            How many units should be trimmed off at the beginning and end of
            the cropping dimension (axis 1).
            If a single int is provided,
            the same value will be used for both.

    Input shape:
        3D tensor with shape `(batch, axis_to_crop, features)`

    Output shape:
        3D tensor with shape `(batch, cropped_axis, features)`

    Polyaxonfile usage:

    ```yaml
    Cropping1D:
    ```

    or

    ```yaml
    Cropping1D:
        cropping: 1 or [1, 1]
    ```
    """
    IDENTIFIER = 'Cropping1D'
    SCHEMA = Cropping1DSchema

    def __init__(self, cropping=(1, 1), **kwargs):
        super(Cropping1DConfig, self).__init__(**kwargs)
        self.cropping = cropping


class Cropping2DSchema(BaseLayerSchema):
    cropping = ObjectOrListObject(ObjectOrListObject(fields.Int, min=2, max=2), min=2, max=2,
                                  default=((0, 0), (0, 0)), missing=((0, 0), (0, 0)))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return Cropping2DConfig


class Cropping2DConfig(BaseLayerConfig):
    """Cropping layer for 2D input (e.g. picture).

    It crops along spatial dimensions, i.e. width and height.

    Args:
        cropping: int, or tuple of 2 ints, or tuple of 2 tuples of 2 ints.
            - If int: the same symmetric cropping
                is applied to width and height.
            - If tuple of 2 ints:
                interpreted as two different
                symmetric cropping values for height and width:
                `(symmetric_height_crop, symmetric_width_crop)`.
            - If tuple of 2 tuples of 2 ints:
                interpreted as
                `((top_crop, bottom_crop), (left_crop, right_crop))`
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, rows, cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, rows, cols)`

    Output shape:
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, cropped_rows, cropped_cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, cropped_rows, cropped_cols)`

    Example:

    ```python
    # Crop the input 2D images or feature maps
    x = Cropping2D(cropping=((2, 2), (4, 4)), input_shape=(28, 28, 3))(x)
    # now x.output_shape == (None, 24, 20, 3)
    x = Conv2D(64, (3, 3), padding='same')(x)
    x = Cropping2D(cropping=((2, 2), (2, 2)))(x)
    # now x.output_shape == (None, 20, 16. 64)
    ```

    Polyaxonfile usage:

    ```yaml
    Cropping2D:
      cropping=[[2, 2], [4, 4]]
    ```
    """
    IDENTIFIER = 'Cropping2D'
    SCHEMA = Cropping2DSchema

    def __init__(self, cropping=((0, 0), (0, 0)), data_format=None, **kwargs):
        super(Cropping2DConfig, self).__init__(**kwargs)
        self.cropping = cropping
        self.data_format = data_format


class Cropping3DSchema(BaseLayerSchema):
    cropping = ObjectOrListObject(ObjectOrListObject(fields.Int, min=2, max=2), min=3, max=3,
                                  default=((1, 1), (1, 1), (1, 1)),
                                  missing=((1, 1), (1, 1), (1, 1)))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return Cropping3DConfig


class Cropping3DConfig(BaseLayerConfig):
    """Cropping layer for 3D data (e.g.

    spatial or spatio-temporal).

    Args:
        cropping: int, or tuple of 23ints, or tuple of 3 tuples of 2 ints.
            - If int: the same symmetric cropping
                is applied to depth, height, and width.
            - If tuple of 3 ints:
                interpreted as two different
                symmetric cropping values for depth, height, and width:
                `(symmetric_dim1_crop, symmetric_dim2_crop, symmetric_dim3_crop)`.
            - If tuple of 3 tuples of 2 ints:
                interpreted as
                `((left_dim1_crop, right_dim1_crop), (left_dim2_crop,
                  right_dim2_crop), (left_dim3_crop, right_dim3_crop))`
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        5D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, first_axis_to_crop, second_axis_to_crop, third_axis_to_crop,
              depth)`
        - If `data_format` is `"channels_first"`:
            `(batch, depth, first_axis_to_crop, second_axis_to_crop,
              third_axis_to_crop)`

    Output shape:
        5D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, first_cropped_axis, second_cropped_axis, third_cropped_axis,
              depth)`
        - If `data_format` is `"channels_first"`:
            `(batch, depth, first_cropped_axis, second_cropped_axis,
              third_cropped_axis)`

    Polyaxonfile usage:

    ```yaml
    Cropping3D:
      cropping=[[2, 2], [4, 4], [2, 2]]
    ```
    """
    IDENTIFIER = 'Cropping3D'
    SCHEMA = Cropping3DSchema

    def __init__(self, cropping=((1, 1), (1, 1), (1, 1)), data_format=None, **kwargs):
        super(Cropping3DConfig, self).__init__(**kwargs)
        self.cropping = cropping
        self.data_format = data_format



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError

from polyaxon_schemas.ops.build import BuildConfig
from polyaxon_schemas.utils import BuildBackend


class TestBuildConfigs(TestCase):
    def test_valid_image(self):
        config_dict = {
            'image': None,
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': '',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'some_image_name:sdf:sdf:foo',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'registry.foobar.com/my/docker/some_image_name:foo:foo',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'some_image_name / foo',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'some_image_name /foo:sdf',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'some_image_name /foo :sdf',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'registry.foobar.com:foo:foo/my/docker/some_image_name:foo',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'registry.foobar.com:foo:foo/my/docker/some_image_name',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'image': 'registry.foobar.com:/my/docker/some_image_name:foo',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

    def test_valid_dockerfile(self):
        config_dict = {
            'dockerfile': None,
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

        config_dict = {
            'dockerfile': '',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

    def test_build_config(self):
        config_dict = {
            'image': 'some_image_name',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == 'latest'
        assert config.nocache is None
        assert config.dockerfile is None
        assert config.context is None

        config_dict = {
            'dockerfile': 'path/Dockerfile',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag is None
        assert config.nocache is None
        assert config.dockerfile == 'path/Dockerfile'
        assert config.context is None

        config_dict = {
            'dockerfile': 'path/Dockerfile',
            'image': 'some_image_name',
        }
        with self.assertRaises(ValidationError):
            BuildConfig.from_dict(config_dict)

    def test_build_config_image_use_cases(self):
        # Latest
        config_dict = {
            'image': 'some_image_name',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == 'latest'

        # Latest from with docker registry url
        config_dict = {
            'image': 'registry.foobar.com/my/docker/some_image_name',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == 'latest'

        # Latest from with docker registry url with port
        config_dict = {
            'image': 'registry.foobar.com:4567/some_image_name',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == 'latest'

        # Some tag
        config_dict = {
            'image': 'some_image_name:4567',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == '4567'

        # Some tag
        config_dict = {
            'image': 'some_image_name:foo',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == 'foo'

        # Some tag from with docker registry url
        config_dict = {
            'image': 'registry.foobar.com/my/docker/some_image_name:foo',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == 'foo'

        # Some tag from with docker registry url with port
        config_dict = {
            'image': 'registry.foobar.com:4567/some_image_name:foo',
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.image_tag == 'foo'

    def test_build_nocache(self):
        config_dict = {
            'image': 'some_image_name',
            'nocache': True
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == 'latest'
        assert config.nocache is True

        config_dict = {
            'dockerfile': 'Dockerfile',
            'nocache': True
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag is None
        assert config.dockerfile == 'Dockerfile'
        assert config.nocache is True
        assert config.backend is None

    def test_build_context(self):
        config_dict = {
            'image': 'some_image_name',
            'context': 'path/to/module'
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == 'latest'
        assert config.context == 'path/to/module'

        config_dict = {
            'dockerfile': 'path/to/Dockerfile',
            'context': 'path/to/module'
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.dockerfile == 'path/to/Dockerfile'
        assert config.context == 'path/to/module'
        assert config.backend is None

    def test_build_repo_with_install_step_config(self):
        config_dict = {
            'image': 'tensorflow:1.3.0',
            'build_steps': ['pip install tensor2tensor'],
            'env_vars': [['LC_ALL', 'en_US.UTF-8']],
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == '1.3.0'
        assert config.backend is None

    def test_build_with_native(self):
        config_dict = {
            'image': 'tensorflow:1.3.0',
            'build_steps': ['pip install tensor2tensor'],
            'env_vars': [['LC_ALL', 'en_US.UTF-8']],
            'backend': BuildBackend.NATIVE
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == '1.3.0'
        assert config.backend == BuildBackend.NATIVE

    def test_build_with_kaniko(self):
        config_dict = {
            'image': 'tensorflow:1.3.0',
            'build_steps': ['pip install tensor2tensor'],
            'env_vars': [['LC_ALL', 'en_US.UTF-8']],
            'backend': BuildBackend.KANIKO
        }
        config = BuildConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.image_tag == '1.3.0'
        assert config.backend == BuildBackend.KANIKO



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from hestia.cached_property import cached_property
from marshmallow import EXCLUDE

from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.ops.build import BuildConfig
from polyaxon_schemas.specs.base import BaseSpecification, EnvironmentSpecificationMixin


class BuildSpecification(BaseSpecification, EnvironmentSpecificationMixin):
    """The polyaxonfile specification for build jobs.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
    """
    _SPEC_KIND = BaseSpecification._BUILD

    BUILD_STEPS = 'build_steps'
    ENV_VARS = 'env_vars'
    NOCACHE = 'nocache'
    BRANCH = 'branch'
    COMMIT = 'commit'
    CONTEXT = 'context'
    DOCKERFILE = 'dockerfile'
    IMAGE = 'image'

    SECTIONS = BaseSpecification.SECTIONS + (
        BUILD_STEPS, ENV_VARS, NOCACHE, BRANCH, COMMIT, CONTEXT, DOCKERFILE, IMAGE)

    OP_PARSING_SECTIONS = BaseSpecification.OP_PARSING_SECTIONS + (
        BUILD_STEPS, ENV_VARS, NOCACHE, BRANCH, COMMIT, CONTEXT, DOCKERFILE, IMAGE)

    HEADER_SECTIONS = BaseSpecification.HEADER_SECTIONS + (BaseSpecification.BACKEND, )

    POSSIBLE_SECTIONS = BaseSpecification.POSSIBLE_SECTIONS + (
        BaseSpecification.BACKEND, BaseSpecification.ENVIRONMENT,
        BUILD_STEPS, ENV_VARS, NOCACHE, BRANCH, COMMIT, CONTEXT, DOCKERFILE, IMAGE
    )

    CONFIG = BuildConfig

    @cached_property
    def backend(self):
        return self.config.backend

    @classmethod
    def create_specification(cls,
                             build_config,
                             secret_refs=None,
                             configmap_refs=None,
                             to_dict=True):
        if isinstance(build_config, BuildConfig):
            config = build_config.to_light_dict()
        elif isinstance(build_config, Mapping):
            # Since the objective is to create the build spec from other specs
            # we drop any extra attrs
            config = BuildConfig.from_dict(build_config, unknown=EXCLUDE)
            config = config.to_light_dict()
        else:
            raise PolyaxonConfigurationError(
                'Create specification expects a dict or an instance of BuildConfig.')

        specification = {
            cls.VERSION: 1,
            cls.KIND: cls._SPEC_KIND,
        }
        specification.update(config)

        env = {}
        if secret_refs:
            env['secret_refs'] = secret_refs
        if configmap_refs:
            env['configmap_refs'] = configmap_refs
        if env:
            specification[cls.ENVIRONMENT] = env

        if to_dict:
            return specification
        return cls.read(specification)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.ops.run import BaseRunConfig, BaseRunSchema
from polyaxon_schemas.ops.run_exec import RunSchema


class JobSchema(BaseRunSchema):
    kind = fields.Str(allow_none=None, validate=validate.Equal('job'))
    declarations = fields.Raw(allow_none=True)
    run = fields.Nested(RunSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return JobConfig


class JobConfig(BaseRunConfig):
    SCHEMA = JobSchema
    IDENTIFIER = 'job'

    def __init__(self,
                 kind=None,
                 version=None,
                 logging=None,
                 tags=None,
                 declarations=None,
                 environment=None,
                 build=None,
                 run=None,
                 ):
        super(JobConfig, self).__init__(kind=kind,
                                        version=version,
                                        logging=logging,
                                        tags=tags,
                                        environment=environment,
                                        build=build)
        self.declarations = declarations
        self.run = run



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.utils import ObjectOrListObject


class MaxPooling1DSchema(BaseLayerSchema):
    pool_size = fields.Int(default=2, missing=2, allow_none=True)
    strides = fields.Int(default=None, missing=None)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))

    @staticmethod
    def schema_config():
        return MaxPooling1DConfig


class MaxPooling1DConfig(BaseLayerConfig):
    """Max pooling operation for temporal data.

    Args:
        pool_size: Integer, size of the max pooling windows.
        strides: Integer, or None. Factor by which to downscale.
            E.g. 2 will halve the input.
            If None, it will default to `pool_size`.
        padding: One of `"valid"` or `"same"` (case-insensitive).

    Input shape:
        3D tensor with shape: `(batch_size, steps, features)`.

    Output shape:
        3D tensor with shape: `(batch_size, downsampled_steps, features)`.

    Polyaxonfile usage:

    ```yaml
    MaxPooling1D:
      pool_size: 2
    ```
    """
    IDENTIFIER = 'MaxPooling1D'
    SCHEMA = MaxPooling1DSchema

    def __init__(self, pool_size=2, strides=None, padding='valid', **kwargs):
        super(MaxPooling1DConfig, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding


class AveragePooling1DSchema(BaseLayerSchema):
    pool_size = fields.Int(default=2, missing=2, allow_none=True)
    strides = fields.Int(default=None, missing=None)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))

    @staticmethod
    def schema_config():
        return AveragePooling1DConfig


class AveragePooling1DConfig(BaseLayerConfig):
    """Average pooling for temporal data.

    Args:
        pool_size: Integer, size of the max pooling windows.
        strides: Integer, or None. Factor by which to downscale.
            E.g. 2 will halve the input.
            If None, it will default to `pool_size`.
        padding: One of `"valid"` or `"same"` (case-insensitive).

    Input shape:
        3D tensor with shape: `(batch_size, steps, features)`.

    Output shape:
        3D tensor with shape: `(batch_size, downsampled_steps, features)`.

    Polyaxonfile usage:

    ```yaml
    AveragePooling1D:
      pool_size: 2
    ```
    """
    IDENTIFIER = 'AveragePooling1D'
    SCHEMA = AveragePooling1DSchema

    def __init__(self, pool_size=2, strides=None, padding='valid', **kwargs):
        super(AveragePooling1DConfig, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding


class MaxPooling2DSchema(BaseLayerSchema):
    pool_size = ObjectOrListObject(fields.Int, min=2, max=2, default=(2, 2), missing=(2, 2))
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=None, missing=None)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return MaxPooling2DConfig


class MaxPooling2DConfig(BaseLayerConfig):
    """Max pooling operation for spatial data.

    Args:
        pool_size: integer or tuple of 2 integers,
            factors by which to downscale (vertical, horizontal).
            (2, 2) will halve the input in both spatial dimension.
            If only one integer is specified, the same window length
            will be used for both dimensions.
        strides: Integer, tuple of 2 integers, or None.
            Strides values.
            If None, it will default to `pool_size`.
        padding: One of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, rows, cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, rows, cols)`

    Output shape:
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, pooled_rows, pooled_cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, pooled_rows, pooled_cols)`

    Polyaxonfile usage:

    ```yaml
    MaxPooling2D:
      pool_size: [2, 2]
    ```
    """
    IDENTIFIER = 'MaxPooling2D'
    SCHEMA = MaxPooling2DSchema

    def __init__(self, pool_size=(2, 2), strides=None, padding='valid', data_format=None, **kwargs):
        super(MaxPooling2DConfig, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format


class AveragePooling2DSchema(BaseLayerSchema):
    pool_size = ObjectOrListObject(fields.Int, min=2, max=2, default=(2, 2), missing=(2, 2))
    strides = ObjectOrListObject(fields.Int, min=2, max=2, default=None, missing=None)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return AveragePooling2DConfig


class AveragePooling2DConfig(BaseLayerConfig):
    """Average pooling operation for spatial data.

    Args:
        pool_size: integer or tuple of 2 integers,
            factors by which to downscale (vertical, horizontal).
            (2, 2) will halve the input in both spatial dimension.
            If only one integer is specified, the same window length
            will be used for both dimensions.
        strides: Integer, tuple of 2 integers, or None.
            Strides values.
            If None, it will default to `pool_size`.
        padding: One of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, rows, cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, rows, cols)`

    Output shape:
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, pooled_rows, pooled_cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, pooled_rows, pooled_cols)`

    Polyaxonfile usage:

    ```yaml
    AveragePooling2D:
      pool_size: [2, 2]
    ```
    """
    IDENTIFIER = 'AveragePooling2D'
    SCHEMA = AveragePooling2DSchema

    def __init__(self, pool_size=(2, 2), strides=None, padding='valid', data_format=None, **kwargs):
        super(AveragePooling2DConfig, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format


class MaxPooling3DSchema(BaseLayerSchema):
    pool_size = ObjectOrListObject(fields.Int, min=3, max=3, default=(2, 2, 2), missing=(2, 2, 2))
    strides = ObjectOrListObject(fields.Int, min=3, max=3, default=None, missing=None)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return MaxPooling3DConfig


class MaxPooling3DConfig(BaseLayerConfig):
    """Max pooling operation for 3D data (spatial or spatio-temporal).

    Args:
        pool_size: tuple of 3 integers,
            factors by which to downscale (dim1, dim2, dim3).
            (2, 2, 2) will halve the size of the 3D input in each dimension.
        strides: tuple of 3 integers, or None. Strides values.
        padding: One of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            5D tensor with shape:
            `(batch_size, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
        - If `data_format='channels_first'`:
            5D tensor with shape:
            `(batch_size, channels, spatial_dim1, spatial_dim2, spatial_dim3)`

    Output shape:
        - If `data_format='channels_last'`:
            5D tensor with shape:
            `(batch_size, pooled_dim1, pooled_dim2, pooled_dim3, channels)`
        - If `data_format='channels_first'`:
            5D tensor with shape:
            `(batch_size, channels, pooled_dim1, pooled_dim2, pooled_dim3)`

    Polyaxonfile usage:

    ```yaml
    MaxPooling3D:
      pool_size: [2, 2, 2]
    ```
    """
    IDENTIFIER = 'MaxPooling3D'
    SCHEMA = MaxPooling3DSchema

    def __init__(self, pool_size=(2, 2, 2), strides=None, padding='valid', data_format=None,
                 **kwargs):
        super(MaxPooling3DConfig, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format


class AveragePooling3DSchema(BaseLayerSchema):
    pool_size = ObjectOrListObject(fields.Int, min=3, max=3, default=(2, 2, 2), missing=(2, 2, 2))
    strides = ObjectOrListObject(fields.Int, min=3, max=3, default=None, missing=None)
    padding = fields.Str(default='valid', missing='valid',
                         validate=validate.OneOf(['same', 'valid']))
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return AveragePooling3DConfig


class AveragePooling3DConfig(BaseLayerConfig):
    """Average pooling operation for 3D data (spatial or spatio-temporal).

    Args:
        pool_size: tuple of 3 integers,
            factors by which to downscale (dim1, dim2, dim3).
            (2, 2, 2) will halve the size of the 3D input in each dimension.
        strides: tuple of 3 integers, or None. Strides values.
        padding: One of `"valid"` or `"same"` (case-insensitive).
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            5D tensor with shape:
            `(batch_size, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
        - If `data_format='channels_first'`:
            5D tensor with shape:
            `(batch_size, channels, spatial_dim1, spatial_dim2, spatial_dim3)`

    Output shape:
        - If `data_format='channels_last'`:
            5D tensor with shape:
            `(batch_size, pooled_dim1, pooled_dim2, pooled_dim3, channels)`
        - If `data_format='channels_first'`:
            5D tensor with shape:
            `(batch_size, channels, pooled_dim1, pooled_dim2, pooled_dim3)`

    Polyaxonfile usage:

    ```yaml
    AveragePooling3D:
      pool_size: [2, 2, 2]
    ```
    """
    IDENTIFIER = 'AveragePooling3D'
    SCHEMA = AveragePooling3DSchema

    def __init__(self, pool_size=(2, 2, 2), strides=None, padding='valid', data_format=None,
                 **kwargs):
        super(AveragePooling3DConfig, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format


class GlobalAveragePooling1DSchema(BaseLayerSchema):
    @staticmethod
    def schema_config():
        return GlobalAveragePooling1DConfig


class GlobalAveragePooling1DConfig(BaseLayerConfig):
    """Global average pooling operation for temporal data.

    Input shape:
        3D tensor with shape: `(batch_size, steps, features)`.

    Output shape:
        2D tensor with shape:
        `(batch_size, channels)`

    Polyaxonfile usage:

    ```yaml
    GlobalAveragePooling1D:
    ```
    """
    IDENTIFIER = 'GlobalAveragePooling1D'
    SCHEMA = GlobalAveragePooling1DSchema


class GlobalMaxPooling1DSchema(BaseLayerSchema):
    @staticmethod
    def schema_config():
        return GlobalMaxPooling1DConfig


class GlobalMaxPooling1DConfig(BaseLayerConfig):
    """Global max pooling operation for temporal data.

    Input shape:
        3D tensor with shape: `(batch_size, steps, features)`.

    Output shape:
        2D tensor with shape:
        `(batch_size, channels)`

    Polyaxonfile usage:

    ```yaml
    GlobalMaxPooling1D:
    ```
    """
    IDENTIFIER = 'GlobalMaxPooling1D'
    SCHEMA = GlobalMaxPooling1DSchema


class GlobalAveragePooling2DSchema(BaseLayerSchema):
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return GlobalAveragePooling2DConfig


class GlobalAveragePooling2DConfig(BaseLayerConfig):
    """Global average pooling operation for spatial data.

    Args:
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, rows, cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, rows, cols)`

    Output shape:
        2D tensor with shape:
        `(batch_size, channels)`

    Polyaxonfile usage:

    ```yaml
    GlobalAveragePooling2D:
    ```
    """
    IDENTIFIER = 'GlobalAveragePooling2D'
    SCHEMA = GlobalAveragePooling2DSchema

    def __init__(self, data_format=None, **kwargs):
        super(GlobalAveragePooling2DConfig, self).__init__(**kwargs)
        self.data_format = data_format


class GlobalMaxPooling2DSchema(BaseLayerSchema):
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return GlobalMaxPooling2DConfig


class GlobalMaxPooling2DConfig(BaseLayerConfig):
    """Global max pooling operation for spatial data.

    Args:
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            4D tensor with shape:
            `(batch_size, rows, cols, channels)`
        - If `data_format='channels_first'`:
            4D tensor with shape:
            `(batch_size, channels, rows, cols)`

    Output shape:
        2D tensor with shape:
        `(batch_size, channels)`

    Polyaxonfile usage:

    ```yaml
    GlobalMaxPooling2D:
    ```
    """
    IDENTIFIER = 'GlobalMaxPooling2D'
    SCHEMA = GlobalMaxPooling2DSchema

    def __init__(self, data_format=None, **kwargs):
        super(GlobalMaxPooling2DConfig, self).__init__(**kwargs)
        self.data_format = data_format


class GlobalAveragePooling3DSchema(BaseLayerSchema):
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return GlobalAveragePooling3DConfig


class GlobalAveragePooling3DConfig(BaseLayerConfig):
    """Global Average pooling operation for 3D data.

    Args:
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            5D tensor with shape:
            `(batch_size, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
        - If `data_format='channels_first'`:
            5D tensor with shape:
            `(batch_size, channels, spatial_dim1, spatial_dim2, spatial_dim3)`

    Output shape:
        2D tensor with shape:
        `(batch_size, channels)`

    Polyaxonfile usage:

    ```yaml
    GlobalAveragePooling3D:
    ```
    """
    IDENTIFIER = 'GlobalAveragePooling3D'
    SCHEMA = GlobalAveragePooling3DSchema

    def __init__(self, data_format=None, **kwargs):
        super(GlobalAveragePooling3DConfig, self).__init__(**kwargs)
        self.data_format = data_format


class GlobalMaxPooling3DSchema(BaseLayerSchema):
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    @staticmethod
    def schema_config():
        return GlobalMaxPooling3DConfig


class GlobalMaxPooling3DConfig(BaseLayerConfig):
    """Global Max pooling operation for 3D data.

    Args:
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
            while `channels_first` corresponds to inputs with shape
            `(batch, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
            If you never set it, then it will be "channels_last".

    Input shape:
        - If `data_format='channels_last'`:
            5D tensor with shape:
            `(batch_size, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
        - If `data_format='channels_first'`:
            5D tensor with shape:
            `(batch_size, channels, spatial_dim1, spatial_dim2, spatial_dim3)`

    Output shape:
        2D tensor with shape:
        `(batch_size, channels)`

    Polyaxonfile usage:

    ```yaml
    GlobalMaxPooling3D:
    ```
    """
    IDENTIFIER = 'GlobalMaxPooling3D'
    SCHEMA = GlobalMaxPooling3DSchema

    def __init__(self, data_format=None, **kwargs):
        super(GlobalMaxPooling3DConfig, self).__init__(**kwargs)
        self.data_format = data_format



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.exceptions import PolyaxonfileError
from polyaxon_schemas.specs import ExperimentSpecification
from polyaxon_schemas.specs.libs.parser import Parser


class TestParser(TestCase):
    def test_parse_base_expressions(self):
        data = [
            1, 10., [1, 1], (1, 1), 'string', ['str1', 'str2'], {1: 2, 'a': 'a', 'dict': {1: 1}}
        ]

        parser = Parser()
        for d in data:
            assert d == parser.parse_expression(ExperimentSpecification, d, {})

    def test_parse_context_expression(self):
        parser = Parser()
        assert parser.parse_expression(ExperimentSpecification, '{{ something }}', {}) == ''
        assert parser.parse_expression(ExperimentSpecification, '{{ something }}',
                                       {'something': 1}) == 1

    def test_parse_graph_expression(self):
        expression = {
            'graph': {
                'input_layers': ['images'],
                'layers': [
                    {'Conv2D': {
                        'filters': 64,
                        'kernel_size': [3, 3],
                        'strides': [1, 1],
                        'activation': 'relu',
                    }},
                    {'MaxPooling2D': {'kernels': 2}},
                    {'Flatten': None},
                    {'Dense': {
                        'units': 10,
                        'activation': 'softmax',
                    }}
                ]
            }
        }

        parser = Parser()
        assert parser.parse_expression(ExperimentSpecification, expression, {}) == expression

        expected_expression = {
            'graph': {'input_layers': ['images'],
                      'layers': [
                          {'Conv2D': {'activation': 'relu',
                                      'filters': 64,
                                      'kernel_size': [3, 3],
                                      'name': 'Conv2D_1',
                                      'strides': [1, 1],
                                      'inbound_nodes': ['images']}
                           },
                          {'MaxPooling2D': {'inbound_nodes': ['Conv2D_1'],
                                            'kernels': 2,
                                            'name': 'MaxPooling2D_1'}
                           },
                          {'Flatten': {'inbound_nodes': ['MaxPooling2D_1'],
                                       'name': 'Flatten_1'}
                           },
                          {'Dense': {'activation': 'softmax',
                                     'inbound_nodes': ['Flatten_1'],
                                     'name': 'Dense_1',
                                     'units': 10}
                           }
                      ],
                      'output_layers': ['Dense_1']}
        }
        assert parser.parse_expression(
            ExperimentSpecification, expression, {}, check_graph=True) == expected_expression

    def test_parse_graph_with_operators_and_tags(self):
        declarations = {
            'conv2d':
                {
                    'filters': [32, 64],
                    'kernel_size': [[3, 3], [2, 2]],
                    'strides': [1, 1],
                    'activation': ['relu', 'linear']
                }
        }
        expression = {
            'graph': {
                'input_layers': ['images'],
                'layers': [
                    {'for': {
                        'len': 2,
                        'index': 'i',
                        'do':
                            {'Conv2D': {
                                'filters': '{{ conv2d.filters[i] }}',
                                'kernel_size': '{{ conv2d.kernel_size[i] }}',
                                'strides': '{{ conv2d.strides }}',
                                'activation': '{{ conv2d.activation[i] }}',
                                'tags': ['tag1', 'tag2']
                            }}
                    }},
                    {'MaxPooling2D': {'kernels': 2}},
                    {'Flatten': {'name': 'Flatten_1'}},
                    {'Flatten': {'inbound_nodes': ['{{ tags.tag1[0] }}'], 'name': 'Flatten_2'}},
                    {'Concat': {'inbound_nodes': ['Flatten_1', 'Flatten_2']}},
                    {'Dense': {
                        'units': 10,
                        'activation': 'softmax',
                    }}
                ]
            }
        }

        parser = Parser()
        result_expression = parser.parse_expression(
            ExperimentSpecification,
            expression,
            declarations,
            check_operators=True,
            check_graph=True)
        expected_result = {'graph': {
            'input_layers': ['images'],
            'layers': [
                {'Conv2D': {'activation': 'relu',
                            'filters': 32,
                            'kernel_size': [3, 3],
                            'name': 'Conv2D_1',
                            'strides': [1, 1],
                            'inbound_nodes': ['images'],
                            'tags': ['tag1', 'tag2']}},
                {'Conv2D': {'activation': 'linear',
                            'filters': 64,
                            'inbound_nodes': ['Conv2D_1'],
                            'kernel_size': [2, 2],
                            'name': 'Conv2D_2',
                            'strides': [1, 1],
                            'tags': ['tag1', 'tag2']}},
                {'MaxPooling2D': {'inbound_nodes': ['Conv2D_2'],
                                  'kernels': 2,
                                  'name': 'MaxPooling2D_1'}},
                {'Flatten': {'inbound_nodes': ['MaxPooling2D_1'],
                             'name': 'Flatten_1'}},
                {'Flatten': {'inbound_nodes': ['Conv2D_1'],
                             'name': 'Flatten_2'}},
                {'Concat': {
                    'inbound_nodes': ['Flatten_1', 'Flatten_2'],
                    'name': 'Concat_1'}},
                {'Dense': {'activation': 'softmax',
                           'inbound_nodes': ['Concat_1'],
                           'name': 'Dense_1',
                           'units': 10}}],
            'output_layers': ['Dense_1']}}
        assert result_expression == expected_result

    def test_parse_graph_with_many_inputs_used(self):
        expression = {
            'graph': {
                'input_layers': ['images'],
                'layers': [
                    {'Conv2D': {
                        'filters': 64,
                        'kernel_size': [3, 3],
                        'strides': [1, 1],
                        'activation': 'relu',
                        'inbound_nodes': ['images']
                    }},
                    {'MaxPooling2D': {'kernels': 2}},
                    {'Flatten': None},
                    {'Dense': {
                        'units': 10,
                        'activation': 'softmax',
                    }}
                ]
            }
        }

        parser = Parser()
        assert parser.parse_expression(ExperimentSpecification, expression, {}) == expression

        expected_expression = {
            'graph': {
                'input_layers': ['images'],
                'layers': [
                    {'Conv2D': {'activation': 'relu',
                                'filters': 64,
                                'kernel_size': [3, 3],
                                'name': 'Conv2D_1',
                                'strides': [1, 1],
                                'inbound_nodes': ['images']}
                     },
                    {'MaxPooling2D': {'inbound_nodes': ['Conv2D_1'],
                                      'kernels': 2,
                                      'name': 'MaxPooling2D_1'}
                     },
                    {'Flatten': {'inbound_nodes': ['MaxPooling2D_1'],
                                 'name': 'Flatten_1'}
                     },
                    {'Dense': {'activation': 'softmax',
                               'inbound_nodes': ['Flatten_1'],
                               'name': 'Dense_1',
                               'units': 10}
                     }
                ],
                'output_layers': ['Dense_1']}
        }
        assert parser.parse_expression(
            ExperimentSpecification, expression, {}, check_graph=True) == expected_expression

    def test_parse_graph_with_many_inputs_and_non_used_raises(self):
        expression = {
            'graph': {
                'input_layers': ['images', 'sounds'],
                'layers': [
                    {'Conv2D': {
                        'filters': 64,
                        'kernel_size': [3, 3],
                        'strides': [1, 1],
                        'activation': 'relu',
                        'name': 'Conv2D_1',
                        'inbound_nodes': ['images']
                    }},
                    {'MaxPooling2D': {'kernels': 2}},  # This layer is orphan
                    {'Flatten': {'inbound_nodes': ['Conv2D_1']}},  # Specify the input layer
                    {'Dense': {
                        'units': 10,
                        'activation': 'softmax',
                    }}
                ]
            }
        }

        parser = Parser()
        with self.assertRaises(PolyaxonfileError):
            parser.parse_expression(ExperimentSpecification, expression, {}, check_graph=True)

    def test_parse_graph_with_many_inputs_and_some_used_raises(self):
        expression = {
            'graph': {
                'input_layers': ['images', 'sounds'],
                'layers': [
                    {'Conv2D': {
                        'filters': 64,
                        'kernel_size': [3, 3],
                        'strides': [1, 1],
                        'activation': 'relu',
                        'name': 'Conv2D_1'
                    }},
                    {'MaxPooling2D': {'kernels': 2}},  # This layer is orphan
                    {'Flatten': {'inbound_nodes': ['Conv2D_1']}},  # Specify the input layer
                    {'Dense': {
                        'units': 10,
                        'activation': 'softmax',
                    }}
                ]
            }
        }

        parser = Parser()
        with self.assertRaises(PolyaxonfileError):
            parser.parse_expression(ExperimentSpecification, expression, {}, check_graph=True)

    def test_parse_graph_with_orphan_layers_raises(self):
        expression = {
            'graph': {
                'input_layers': ['images'],
                'layers': [
                    {'Conv2D': {
                        'filters': 64,
                        'kernel_size': [3, 3],
                        'strides': [1, 1],
                        'activation': 'relu',
                        'name': 'Conv2D_1'
                    }},
                    {'MaxPooling2D': {'kernels': 2}},  # This layer is orphan
                    {'Flatten': {'inbound_nodes': ['Conv2D_1']}},  # Specify the input layer
                    {'Dense': {
                        'units': 10,
                        'activation': 'softmax',
                    }}
                ]
            }
        }

        parser = Parser()
        with self.assertRaises(PolyaxonfileError):
            parser.parse_expression(ExperimentSpecification, expression, {}, check_graph=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.environments.experiments import TensorflowClusterSchema
from polyaxon_schemas.ops.environments.pods import PodEnvironmentConfig, PodEnvironmentSchema


class GPUOptionsSchema(BaseSchema):
    gpu_memory_fraction = fields.Float(allow_none=True)
    allow_growth = fields.Bool(allow_none=True)
    per_process_gpu_memory_fraction = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return GPUOptionsConfig


class GPUOptionsConfig(BaseConfig):
    IDENTIFIER = 'session'
    SCHEMA = GPUOptionsSchema

    def __init__(self,
                 gpu_memory_fraction=None,
                 allow_growth=True,
                 per_process_gpu_memory_fraction=None):
        self.gpu_memory_fraction = gpu_memory_fraction
        self.allow_growth = allow_growth
        self.per_process_gpu_memory_fraction = per_process_gpu_memory_fraction


class SessionSchema(BaseSchema):
    gpu_options = fields.Nested(GPUOptionsSchema, allow_none=True)
    log_device_placement = fields.Bool(allow_none=True)
    allow_soft_placement = fields.Bool(allow_none=True)
    intra_op_parallelism_threads = fields.Int(allow_none=True)
    inter_op_parallelism_threads = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return SessionConfig


class SessionConfig(BaseConfig):
    IDENTIFIER = 'session'
    SCHEMA = SessionSchema

    def __init__(self,
                 log_device_placement=True,
                 gpu_options=GPUOptionsConfig(),
                 allow_soft_placement=True,
                 intra_op_parallelism_threads=None,
                 inter_op_parallelism_threads=None):
        self.gpu_options = gpu_options
        self.log_device_placement = log_device_placement
        self.allow_soft_placement = allow_soft_placement
        self.intra_op_parallelism_threads = intra_op_parallelism_threads
        self.inter_op_parallelism_threads = inter_op_parallelism_threads


class TFRunSchema(BaseSchema):
    tf_random_seed = fields.Int(allow_none=True)
    save_summary_steps = fields.Int(allow_none=True)
    save_checkpoints_secs = fields.Int(allow_none=True)
    save_checkpoints_steps = fields.Int(allow_none=True)
    keep_checkpoint_max = fields.Int(allow_none=True)
    keep_checkpoint_every_n_hours = fields.Int(allow_none=True)

    session = fields.Nested(SessionSchema, allow_none=True)
    cluster = fields.Nested(TensorflowClusterSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return TFRunConfig


class TFRunConfig(BaseConfig):
    IDENTIFIER = 'run'
    SCHEMA = TFRunSchema

    def __init__(self,
                 tf_random_seed=None,
                 save_summary_steps=100,
                 save_checkpoints_secs=None,
                 save_checkpoints_steps=None,
                 keep_checkpoint_max=5,
                 keep_checkpoint_every_n_hours=10000,
                 session=None,
                 cluster=None):
        self.tf_random_seed = tf_random_seed
        self.save_summary_steps = save_summary_steps
        self.save_checkpoints_secs = save_checkpoints_secs
        self.save_checkpoints_steps = save_checkpoints_steps
        self.keep_checkpoint_max = keep_checkpoint_max
        self.keep_checkpoint_every_n_hours = keep_checkpoint_every_n_hours
        self.session = session
        self.cluster = cluster


class TensorflowPodEnvironmentSchema(PodEnvironmentSchema):
    config = fields.Nested(SessionSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return TensorflowPodEnvironmentConfig


class TensorflowPodEnvironmentConfig(PodEnvironmentConfig):
    IDENTIFIER = 'pod_environment'
    SCHEMA = TensorflowPodEnvironmentSchema
    REDUCED_ATTRIBUTES = PodEnvironmentConfig.REDUCED_ATTRIBUTES + ['config']

    def __init__(self,
                 index=None,
                 config=None,
                 resources=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 ):
        self.config = config
        super(TensorflowPodEnvironmentConfig, self).__init__(
            index=index,
            resources=resources,
            node_selector=node_selector,
            affinity=affinity,
            tolerations=tolerations,
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.utils import UUID


class DataDetailsSchema(BaseSchema):
    state = fields.Str()
    size = fields.Float()
    uri = fields.Url()

    @staticmethod
    def schema_config():
        return DataDetailsConfig


class DataDetailsConfig(BaseConfig):
    SCHEMA = DataDetailsSchema
    IDENTIFIER = 'data_details'

    def __init__(self, state, size, uri):
        self.state = state
        self.size = size
        self.uri = uri


class DataSchema(BaseSchema):
    uuid = UUID()
    name = fields.Str()
    created_at = fields.LocalDateTime()
    description = fields.Str(allow_none=True)
    details = fields.Nested(DataDetailsSchema)
    version = fields.Str(allow_none=True)
    resource_id = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return DataConfig


class DataConfig(BaseConfig):
    SCHEMA = DataSchema
    IDENTIFIER = 'data'
    DATETIME_ATTRIBUTES = ['created_at']

    def __init__(self,
                 uuid,
                 name,
                 created_at,
                 description,
                 details,
                 version=None,
                 resource_id=None):
        self.uuid = uuid
        self.name = name
        self.created_at = self.localize_date(created_at)
        self.description = description
        self.details = details
        self.version = int(float(version)) if version else None
        self.resource_id = resource_id


class DatasetSchema(BaseSchema):
    uuid = UUID()
    name = fields.Str()
    description = fields.Str(allow_none=True)
    is_public = fields.Boolean()

    @staticmethod
    def schema_config():
        return DatasetConfig


class DatasetConfig(BaseConfig):
    SCHEMA = DatasetSchema
    IDENTIFIER = 'dataset'

    def __init__(self, uuid, name, description, is_public):
        self.uuid = uuid
        self.name = name
        self.description = description
        self.is_public = is_public



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import uuid

from unittest import TestCase

from polyaxon_schemas.api.job import JobLabelConfig


class TestSettingConfigs(TestCase):
    @staticmethod
    def create_pod_labels():
        project_name = 'user/test-id1'
        experiment_group_name = 'user/test-id1/1'
        experiment_name = 'user/test-id1/1/2'
        project_uuid = uuid.uuid4().hex
        experiment_group_uuid = uuid.uuid4().hex
        experiment_uuid = uuid.uuid4().hex
        job_uuid = uuid.uuid4().hex
        task_type = 'master'
        return {'project_name': project_name,
                'experiment_group_name': experiment_group_name,
                'experiment_name': experiment_name,
                'project_uuid': project_uuid,
                'experiment_group_uuid': experiment_group_uuid,
                'experiment_uuid': experiment_uuid,
                'task_type': task_type,
                'task_idx': '0',
                'job_uuid': job_uuid,
                'role': 'polyaxon-worker',
                'type': 'polyaxon-experiment'
                }

    def test_pod_label_config(self):
        config_dict = self.create_pod_labels()
        config = JobLabelConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.initializations import (
    GlorotNormalInitializerConfig,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.core import (
    ActivationConfig,
    ActivityRegularizationConfig,
    CastConfig,
    DenseConfig,
    DropoutConfig,
    FlattenConfig,
    MaskingConfig,
    PermuteConfig,
    RepeatVectorConfig,
    ReshapeConfig,
    SpatialDropout1DConfig,
    SpatialDropout2DConfig,
    SpatialDropout3DConfig
)


class TestCoreConfigs(TestCase):
    def test_masking_config(self):
        config_dict = {
            'mask_value': 0.
        }
        config = MaskingConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    @staticmethod
    def assert_dropout_config(dropout_class, dim):
        config_dict = {
            'rate': 0.,
            'noise_shape': [1, 1],
            'seed': None
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = dropout_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_dropout_config(self):
        self.assert_dropout_config(DropoutConfig, 1)
        self.assert_dropout_config(SpatialDropout1DConfig, 1)
        self.assert_dropout_config(SpatialDropout2DConfig, 2)
        self.assert_dropout_config(SpatialDropout3DConfig, 3)

    def test_activation_config(self):
        config_dict = {
            'activation': 'relu',
        }
        config = ActivationConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_reshape_config(self):
        config_dict = {
            'target_shape': [1, 1]
        }
        config = ReshapeConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_permute_config(self):
        config_dict = {
            'dims': [1, 1]
        }
        config = PermuteConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_flatten_config(self):
        config_dict = {}
        config = FlattenConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_repeat_vector_config(self):
        config_dict = {
            'n': 12,
        }
        config = RepeatVectorConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_dense_config(self):
        config_dict = {
            'units': 12,
            'activation': 'elu',
            'use_bias': True,
            'kernel_initializer': GlorotNormalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'kernel_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'kernel_constraint': None,
            'bias_constraint': None,
            'inbound_nodes': [['layer_1', 0, 1], ['layer_2', 1, 1]]
        }
        config = DenseConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_activity_regularization_config(self):
        config_dict = {
            'l1': 0.2,
            'l2': 0.32,
        }
        config = ActivityRegularizationConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_cast_config(self):
        config_dict = {
            'dtype': 'float32',
        }
        config = CastConfig.from_dict(config_dict)
        assert_equal_layers(config, config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from hestia.cached_property import cached_property

from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.ops.build import BuildConfig
from polyaxon_schemas.ops.environments.persistence import PersistenceConfig
from polyaxon_schemas.specs.base import BaseSpecification
from polyaxon_schemas.specs.experiment import ExperimentSpecification
from polyaxon_schemas.specs.libs import validator
from polyaxon_schemas.specs.libs.parser import Parser
from polyaxon_schemas.utils import SearchAlgorithms


class GroupSpecification(BaseSpecification):
    """Parses Polyaxonfiles/Configuration, with matrix section definition.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        HYPER_PARAMS: hyper params tuning and concurrent runs.
        ENVIRONMENT: defines the run environment for experiment.
        DECLARATIONS: variables/modules that can be reused.
        RUN: defines the run step where the user can set a docker image to execute
        MODEL: defines the model to use based on the declarative API.
        TRAIN: defines how to train a model and how to read the data.
        EVAL: defines how to evaluate a model and how to read the data.
    """

    _SPEC_KIND = BaseSpecification._GROUP

    MODEL = 'model'
    TRAIN = 'train'
    EVAL = 'eval'

    SECTIONS = ExperimentSpecification.SECTIONS + (
        BaseSpecification.HP_TUNING,
    )

    HEADER_SECTIONS = ExperimentSpecification.HEADER_SECTIONS + (
        BaseSpecification.HP_TUNING,
    )

    GRAPH_SECTIONS = ExperimentSpecification.GRAPH_SECTIONS

    REQUIRED_SECTIONS = ExperimentSpecification.REQUIRED_SECTIONS + (
        BaseSpecification.HP_TUNING,
    )

    POSSIBLE_SECTIONS = ExperimentSpecification.POSSIBLE_SECTIONS + (
        BaseSpecification.HP_TUNING,
    )

    def _extra_validation(self):
        if not self.matrix:
            raise PolyaxonConfigurationError(
                'A matrix definition is required for group specification.')

    def _set_parsed_data(self):
        # We need to validate that the data is correct
        # For that we just use a matrix declaration test
        parsed_data = Parser.parse(self, self._data, self.matrix_declaration_test)
        validator.validate(spec=self, data=parsed_data)

    def get_experiment_spec(self, matrix_declaration):
        """Returns an experiment spec for this group spec and the given matrix declaration."""
        parsed_data = Parser.parse(self, self._data, matrix_declaration)
        del parsed_data[self.HP_TUNING]
        validator.validate(spec=self, data=parsed_data)
        return ExperimentSpecification(values=[parsed_data, {'kind': self._EXPERIMENT}])

    def get_build_spec(self):
        """Returns a build spec for this group spec."""
        if BaseSpecification.BUILD not in self._data:
            return None
        return BuildConfig.from_dict(self._data[BaseSpecification.BUILD])

    @cached_property
    def build(self):
        build_spec = self.get_build_spec()
        return self.get_build_spec() if build_spec else None

    @cached_property
    def environment(self):
        # This is a hack, in the future we need to gather the paths of the experiments
        parsed_data = Parser.parse(self, self._data, self.matrix_declaration_test)
        return parsed_data.get(self.ENVIRONMENT, None)

    @cached_property
    def secret_refs(self):
        return self.environment.get('secret_refs') if self.environment else None

    @cached_property
    def configmap_refs(self):
        return self.environment.get('configmap_refs') if self.environment else None

    @cached_property
    def persistence(self):
        persistence = self.environment.get('persistence') if self.environment else None
        return PersistenceConfig.from_dict(persistence) if persistence else None

    @cached_property
    def hptuning(self):
        return self.headers[self.HP_TUNING]

    @cached_property
    def matrix(self):
        if self.hptuning:
            return self.hptuning.matrix
        return None

    @cached_property
    def matrix_space(self):
        if not self.matrix:
            return 1

        space_size = 1

        for value in six.itervalues(self.matrix):
            space_size *= len(value.to_numpy())
        return space_size

    @cached_property
    def early_stopping(self):
        early_stopping = None
        if self.hptuning:
            early_stopping = self.hptuning.early_stopping
        return early_stopping or []

    @cached_property
    def search_algorithm(self):
        return self.hptuning.search_algorithm

    @cached_property
    def concurrency(self):
        concurrency = None
        if self.hptuning:
            concurrency = self.hptuning.concurrency
        return concurrency or 1

    @cached_property
    def experiments_def(self):
        definition = {
            'search_algorithm': self.search_algorithm,
            'early_stopping': True if self.early_stopping else False,
            'concurrency': self.concurrency
        }
        if SearchAlgorithms.is_grid(self.search_algorithm):
            if self.hptuning.grid_search and self.hptuning.grid_search.n_experiments:
                definition['n_experiments'] = self.hptuning.grid_search.n_experiments
        if SearchAlgorithms.is_random(self.search_algorithm):
            if self.hptuning.random_search and self.hptuning.random_search.n_experiments:
                definition['n_experiments'] = self.hptuning.random_search.n_experiments

        return definition

    @cached_property
    def matrix_declaration_test(self):
        if not self.matrix:
            return {}

        return {k: v.sample() for k, v in six.iteritems(self.matrix)}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError
from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.environments.experiments import (
    HorovodClusterConfig,
    HorovodConfig,
    MPIClusterConfig,
    MXNetClusterConfig,
    MXNetConfig,
    PytorchClusterConfig,
    PytorchConfig,
    TensorflowClusterConfig,
    TensorflowConfig
)
from polyaxon_schemas.ops.environments.legacy import TFRunConfig
from polyaxon_schemas.ops.environments.resources import K8SResourcesConfig, PodResourcesConfig


class TestExperimentEnvironmentsConfigs(TestCase):

    def test_tensorflow_cluster_config(self):
        config_dict = {
            "worker": [
                "worker0.example.com:2222",
                "worker1.example.com:2222",
                "worker2.example.com:2222"
            ],
            "ps": [
                "ps0.example.com:2222",
                "ps1.example.com:2222"
            ]
        }
        config = TensorflowClusterConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_horovod_cluster_config(self):
        config_dict = {
            "worker": [
                "worker0.example.com:2222",
                "worker1.example.com:2222",
                "worker2.example.com:2222"
            ]
        }
        config = HorovodClusterConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_mpi_cluster_config(self):
        config_dict = {
            "worker": [
                "worker0.example.com:2222",
                "worker1.example.com:2222",
                "worker2.example.com:2222"
            ]
        }
        config = MPIClusterConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_pytorch_cluster_config(self):
        config_dict = {
            "worker": [
                "worker0.example.com:2222",
                "worker1.example.com:2222",
                "worker2.example.com:2222"
            ]
        }
        config = PytorchClusterConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_mxnet_cluster_config(self):
        config_dict = {
            "worker": [
                "worker0.example.com:2222",
                "worker1.example.com:2222",
                "worker2.example.com:2222"
            ],
            "server": [
                "server0.example.com:2222",
                "server1.example.com:2222"
            ]
        }
        config = MXNetClusterConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_tensorflow_config(self):
        config_dict = {
            'n_workers': 10,
            'n_ps': 5,
        }
        config = TensorflowConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add run config
        config_dict['run_config'] = TFRunConfig().to_dict()
        with self.assertRaises(ValidationError):
            TensorflowConfig.from_dict(config_dict)
        del config_dict['run_config']

        # Add default worker resources
        config_dict['default_worker'] = {
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                gpu=K8SResourcesConfig(2, 4),
                tpu=K8SResourcesConfig(2, 8)
            ).to_dict(),
        }
        config = TensorflowConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_light_dict())

        # Add default ps resources
        config_dict['default_ps'] = {
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400)).to_dict()
        }
        config = TensorflowConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_light_dict())

        # Adding custom resources for worker 4
        config_dict['worker'] = [{
            'index': 4,
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400),
            ).to_dict()
        }]
        config = TensorflowConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_light_dict())

        # Adding custom resources for ps 4
        config_dict['ps'] = [{
            'index': 4,
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400)).to_dict()
        }]
        config = TensorflowConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_light_dict())

    def test_horovod_config(self):
        config_dict = {
            'n_workers': 10,
        }
        config = HorovodConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_light_dict())

        # Add default worker resources
        config_dict['default_worker'] = {
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                tpu=K8SResourcesConfig(2, 8),
                gpu=K8SResourcesConfig(2, 4)).to_dict()
        }
        config = HorovodConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Adding custom resources for worker 4
        config_dict['worker'] = [{
            'index': 4,
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400)).to_dict()
        }]
        config = HorovodConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_light_dict())

    def test_pytorch_config(self):
        config_dict = {
            'n_workers': 10,
        }
        config = PytorchConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add default worker resources
        config_dict['default_worker'] = {
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                tpu=K8SResourcesConfig(1, 1),
                gpu=K8SResourcesConfig(2, 4)).to_dict()
        }
        config = PytorchConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Adding custom resources for worker 4
        config_dict['worker'] = [{
            'index': 4,
            'resources': PodResourcesConfig(cpu=K8SResourcesConfig(0.5, 1),
                                            memory=K8SResourcesConfig(256, 400)).to_dict()
        }]
        config = PytorchConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

    def test_mxnet_config(self):
        config_dict = {
            'n_workers': 10,
            'n_ps': 5,
        }
        config = MXNetConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add default worker resources
        config_dict['default_worker'] = {
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                tpu=K8SResourcesConfig(1, 1),
                gpu=K8SResourcesConfig(2, 4)).to_dict()
        }
        config = MXNetConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Add default ps resources
        config_dict['default_ps'] = {
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400)).to_dict()
        }
        config = MXNetConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Adding custom resources for worker 4
        config_dict['worker'] = [{
            'index': 4,
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400)).to_dict()
        }]
        config = MXNetConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())

        # Adding custom resources for ps 4
        config_dict['ps'] = [{
            'index': 4,
            'resources': PodResourcesConfig(
                cpu=K8SResourcesConfig(0.5, 1),
                memory=K8SResourcesConfig(256, 400)).to_dict()
        }]
        config = MXNetConfig.from_dict(config_dict)
        assert_equal_dict(config_dict, config.to_dict())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.ops.experiment import ExperimentConfig, ExperimentSchema
from polyaxon_schemas.ops.hptuning import HPTuningSchema
from polyaxon_schemas.ops.run import BaseRunConfig


class ExperimentGroupSchema(ExperimentSchema):
    hptuning = fields.Nested(HPTuningSchema, allow_none=None)

    @staticmethod
    def schema_config():
        return ExperimentGroupConfig


class ExperimentGroupConfig(ExperimentConfig):
    SCHEMA = ExperimentSchema
    IDENTIFIER = 'experiment'
    REDUCED_ATTRIBUTES = BaseRunConfig.REDUCED_ATTRIBUTES + ['backend', 'framework']

    def __init__(self,
                 kind=None,
                 version=None,
                 logging=None,
                 tags=None,
                 declarations=None,
                 hptuning=None,
                 environment=None,
                 build=None,
                 backend=None,
                 framework=None,
                 run=None,
                 model=None,
                 train=None,
                 eval=None,  # pylint:disable=redefined-builtin
                 ):
        super(ExperimentGroupConfig, self).__init__(
            kind=kind,
            version=version,
            logging=logging,
            tags=tags,
            declarations=declarations,
            environment=environment,
            build=build,
            backend=backend,
            framework=framework,
            run=run,
            model=model,
            train=train,
            eval=eval)
        self.hptuning = hptuning



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema


class AccessTokenSchema(BaseSchema):
    username = fields.Str()
    token = fields.Str()

    @staticmethod
    def schema_config():
        return AccessTokenConfig


class AccessTokenConfig(BaseConfig):
    """
    Access token config.


    Args:
        username: `str`. The user's username.
        token: `str`. The user's token.
    """
    SCHEMA = AccessTokenSchema
    IDENTIFIER = 'token'

    def __init__(self, username, token):
        self.username = username
        self.token = token


class CredentialsSchema(BaseSchema):
    username = fields.Str()
    password = fields.Str()

    @staticmethod
    def schema_config():
        return CredentialsConfig


class CredentialsConfig(BaseConfig):
    """
    Credentials config.


    Args:
        username: `str`. The user's username.
        password: `str`. The user's password.
    """
    SCHEMA = CredentialsSchema
    IDENTIFIER = 'credentials'

    def __init__(self, username, password):
        self.username = username
        self.password = password



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from hestia.humanize import humanize_timedelta
from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.environments.resources import PodResourcesSchema
from polyaxon_schemas.utils import UUID


class ExperimentJobSchema(BaseSchema):
    id = fields.Int(allow_none=True)
    uuid = UUID()
    unique_name = fields.Str(allow_none=True)
    pod_id = fields.Str(allow_none=True)
    role = fields.Str(allow_none=True)
    experiment = fields.Int()
    experiment_name = fields.Str()
    last_status = fields.Str(allow_none=True)
    created_at = fields.LocalDateTime()
    updated_at = fields.LocalDateTime()
    started_at = fields.LocalDateTime(allow_none=True)
    finished_at = fields.LocalDateTime(allow_none=True)
    total_run = fields.Str(allow_none=True)
    resources = fields.Nested(PodResourcesSchema, allow_none=True)
    definition = fields.Dict(allow_none=True)

    @staticmethod
    def schema_config():
        return ExperimentJobConfig


class ExperimentJobConfig(BaseConfig):
    SCHEMA = ExperimentJobSchema
    IDENTIFIER = 'ExperimentJob'
    DEFAULT_EXCLUDE_ATTRIBUTES = [
        'uuid', 'definition', 'experiment', 'unique_name', 'updated_at', 'resources']
    DATETIME_ATTRIBUTES = ['created_at', 'updated_at', 'started_at', 'finished_at']

    def __init__(self,
                 uuid,
                 experiment,
                 created_at,
                 updated_at,
                 definition=None,
                 unique_name=None,
                 pod_id=None,
                 id=None,  # pylint:disable=redefined-builtin
                 role=None,
                 last_status=None,
                 started_at=None,
                 finished_at=None,
                 resources=None,
                 total_run=None):
        self.uuid = uuid
        self.unique_name = unique_name
        self.pod_id = pod_id
        self.id = id
        self.role = role
        self.experiment = experiment
        self.created_at = self.localize_date(created_at)
        self.updated_at = self.localize_date(updated_at)
        self.started_at = self.localize_date(started_at)
        self.finished_at = self.localize_date(finished_at)
        self.definition = definition
        self.last_status = last_status
        self.resources = resources
        self.total_run = total_run
        if all([self.started_at, self.finished_at]):
            self.total_run = humanize_timedelta((self.finished_at - self.started_at).seconds)


class ExperimentSchema(BaseSchema):
    id = fields.Int(allow_none=True)
    uuid = UUID(allow_none=True)
    unique_name = fields.Str(allow_none=True)
    user = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    name = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    project = fields.Str(allow_none=True)
    experiment_group = fields.Str(allow_none=True)
    build_job = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    last_status = fields.Str(allow_none=True)
    last_metric = fields.Dict(allow_none=True)
    backend = fields.Str(allow_none=True)
    framework = fields.Str(allow_none=True)
    created_at = fields.LocalDateTime(allow_none=True)
    updated_at = fields.LocalDateTime(allow_none=True)
    started_at = fields.LocalDateTime(allow_none=True)
    finished_at = fields.LocalDateTime(allow_none=True)
    total_run = fields.Str(allow_none=True)
    is_clone = fields.Bool(allow_none=True)
    has_tensorboard = fields.Bool(allow_none=True)
    config = fields.Dict(allow_none=True)
    num_jobs = fields.Int(allow_none=True)
    declarations = fields.Dict(allow_none=True)
    tags = fields.List(fields.Str(), allow_none=True)
    resources = fields.Nested(PodResourcesSchema, allow_none=True)
    run_env = fields.Dict(allow_none=True)
    in_cluster = fields.Bool(allow_none=True)
    ttl = fields.Int(allow_none=True)
    jobs = fields.Nested(ExperimentJobSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return ExperimentConfig


class ExperimentConfig(BaseConfig):
    SCHEMA = ExperimentSchema
    IDENTIFIER = 'Experiment'
    DEFAULT_INCLUDE_ATTRIBUTES = [
        'id', 'unique_name', 'user', 'experiment_group', 'build_job', 'last_status',
        'created_at', 'started_at', 'finished_at', 'total_run'
    ]
    DATETIME_ATTRIBUTES = ['created_at', 'updated_at', 'started_at', 'finished_at']

    def __init__(self,
                 id=None,  # pylint:disable=redefined-builtin
                 user=None,
                 uuid=None,
                 name=None,
                 unique_name=None,
                 project=None,
                 experiment_group=None,
                 build_job=None,
                 description=None,
                 last_status=None,
                 last_metric=None,
                 created_at=None,
                 updated_at=None,
                 started_at=None,
                 finished_at=None,
                 is_clone=None,
                 has_tensorboard=False,
                 config=None,
                 num_jobs=0,
                 declarations=None,
                 tags=None,
                 resources=None,
                 in_cluster=None,
                 run_env=None,
                 jobs=None,
                 ttl=None,
                 backend=None,
                 framework=None,
                 total_run=None):
        self.id = id
        self.user = user
        self.uuid = uuid
        self.name = name
        self.unique_name = unique_name
        self.project = project
        self.experiment_group = experiment_group
        self.build_job = build_job
        self.description = description
        self.last_status = last_status
        self.last_metric = last_metric
        self.started_at = self.localize_date(started_at)
        self.finished_at = self.localize_date(finished_at)
        self.created_at = self.localize_date(created_at)
        self.updated_at = self.localize_date(updated_at)
        self.is_clone = is_clone
        self.has_tensorboard = has_tensorboard
        self.config = config  # The json compiled content of this experiment
        self.num_jobs = num_jobs
        self.declarations = declarations
        self.tags = tags
        self.resources = resources
        self.in_cluster = in_cluster
        self.run_env = run_env
        self.jobs = jobs
        self.ttl = ttl
        self.backend = backend
        self.framework = framework
        self.total_run = None
        if all([self.started_at, self.finished_at]):
            self.total_run = humanize_timedelta((self.finished_at - self.started_at).seconds)


class ExperimentStatusSchema(BaseSchema):
    id = fields.Int()
    uuid = UUID()
    experiment = fields.Int()
    created_at = fields.LocalDateTime()
    status = fields.Str()
    message = fields.Str(allow_none=True)
    traceback = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return ExperimentStatusConfig


class ExperimentStatusConfig(BaseConfig):
    SCHEMA = ExperimentStatusSchema
    IDENTIFIER = 'ExperimentStatus'
    DATETIME_ATTRIBUTES = ['created_at']
    DEFAULT_EXCLUDE_ATTRIBUTES = ['experiment', 'uuid', 'traceback']

    def __init__(self,
                 id,  # pylint:disable=redefined-builtin
                 uuid,
                 experiment,
                 created_at,
                 status,
                 message=None,
                 traceback=None):
        self.id = id
        self.uuid = uuid
        self.experiment = experiment
        self.created_at = self.localize_date(created_at)
        self.status = status
        self.message = message
        self.traceback = traceback


class ExperimentMetricSchema(BaseSchema):
    id = fields.Int()
    uuid = UUID()
    experiment = fields.Int()
    created_at = fields.LocalDateTime()
    values = fields.Dict()

    @staticmethod
    def schema_config():
        return ExperimentMetricConfig


class ExperimentMetricConfig(BaseConfig):
    SCHEMA = ExperimentMetricSchema
    IDENTIFIER = 'ExperimentMetric'
    DATETIME_ATTRIBUTES = ['created_at']
    DEFAULT_EXCLUDE_ATTRIBUTES = ['experiment', 'uuid']

    def __init__(self,
                 id,  # pylint:disable=redefined-builtin
                 uuid,
                 experiment,
                 created_at,
                 values):
        self.id = id
        self.uuid = uuid
        self.experiment = experiment
        self.created_at = self.localize_date(created_at)
        self.values = values


class ExperimentJobStatusSchema(BaseSchema):
    id = fields.Int()
    uuid = UUID()
    job = fields.Int()
    created_at = fields.LocalDateTime()
    status = fields.Str()
    message = fields.Str(allow_none=True)
    details = fields.Dict(allow_none=True)

    @staticmethod
    def schema_config():
        return ExperimentJobStatusConfig


class ExperimentJobStatusConfig(BaseConfig):
    SCHEMA = ExperimentJobStatusSchema
    IDENTIFIER = 'ExperimentJobStatus'
    DEFAULT_EXCLUDE_ATTRIBUTES = ['job', 'details', 'uuid']
    DATETIME_ATTRIBUTES = ['created_at']

    def __init__(self,
                 id,  # pylint:disable=redefined-builtin
                 uuid,
                 job,
                 created_at,
                 status,
                 message=None,
                 details=None):
        self.id = id
        self.uuid = uuid
        self.job = job
        self.created_at = self.localize_date(created_at)
        self.status = status
        self.message = message
        self.details = details


class ContainerGPUResourcesSchema(BaseSchema):
    index = fields.Int()
    uuid = fields.Str()
    name = fields.Str()
    minor = fields.Int()
    bus_id = fields.Str()
    serial = fields.Str()
    temperature_gpu = fields.Int()
    utilization_gpu = fields.Int()
    power_draw = fields.Int()
    power_limit = fields.Int()
    memory_free = fields.Int()
    memory_used = fields.Int()
    memory_total = fields.Int()
    memory_utilization = fields.Int()
    processes = fields.List(fields.Dict(), allow_none=True)

    @staticmethod
    def schema_config():
        return ContainerGPUResourcesConfig


class ContainerGPUResourcesConfig(BaseConfig):
    SCHEMA = ContainerGPUResourcesSchema
    IDENTIFIER = 'ContainerGPUResources'
    MEM_SIZE_ATTRIBUTES = ['memory_free', 'memory_used', 'memory_total']

    def __init__(self,
                 index,
                 uuid,
                 name,
                 minor,
                 bus_id,
                 serial,
                 temperature_gpu,
                 utilization_gpu,
                 power_draw,
                 power_limit,
                 memory_free,
                 memory_used,
                 memory_total,
                 memory_utilization,
                 processes=None):
        self.index = index
        self.uuid = uuid
        self.name = name
        self.minor = minor
        self.bus_id = bus_id
        self.serial = serial
        self.temperature_gpu = temperature_gpu
        self.utilization_gpu = utilization_gpu
        self.power_draw = power_draw
        self.power_limit = power_limit
        self.memory_free = memory_free
        self.memory_used = memory_used
        self.memory_total = memory_total
        self.memory_utilization = memory_utilization
        self.processes = processes


class ContainerResourcesSchema(BaseSchema):
    job_uuid = UUID()
    experiment_uuid = UUID()
    job_name = fields.Str()
    container_id = fields.Str()
    n_cpus = fields.Int()
    cpu_percentage = fields.Float()
    percpu_percentage = fields.List(fields.Float(), allow_none=True)
    memory_used = fields.Int()
    memory_limit = fields.Int()
    gpu_resources = fields.Nested(ContainerGPUResourcesSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return ContainerResourcesConfig


class ContainerResourcesConfig(BaseConfig):
    SCHEMA = ContainerResourcesSchema
    IDENTIFIER = 'ContainerResources'
    PERCENT_ATTRIBUTES = ['cpu_percentage']
    MEM_SIZE_ATTRIBUTES = ['memory_used', 'memory_limit']

    def __init__(self,
                 job_uuid,
                 experiment_uuid,
                 job_name,
                 container_id,
                 n_cpus,
                 cpu_percentage,
                 percpu_percentage,
                 memory_used,
                 memory_limit,
                 gpu_resources=None):
        self.job_uuid = job_uuid
        self.experiment_uuid = experiment_uuid
        self.job_name = job_name
        self.container_id = container_id
        self.n_cpus = n_cpus
        self.cpu_percentage = cpu_percentage
        self.percpu_percentage = percpu_percentage
        self.memory_used = memory_used
        self.memory_limit = memory_limit
        self.gpu_resources = gpu_resources
